from datetime import datetime

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr
from vbcore import json
from vbcore.datastruct import IntEnum, ObjectDict
from vbcore.db.mixins import CatalogMixin, StandardMixin
from vbcore.db.sqla import LoaderModel


class PaymentStatusValue(IntEnum):
    WAIT = 1
    SUCCESS = 2
    ERROR = 3


class PaymentStatus(CatalogMixin, LoaderModel):
    __abstract__ = True

    values = PaymentStatusValue.to_list()

    def to_dict(self, **__):
        return ObjectDict(id=self.id, stato=self.label)


class Payment(StandardMixin):
    _status_model = PaymentStatus

    request_id = sa.Column(sa.String(255), unique=True, nullable=False)
    response_id = sa.Column(sa.String(255), unique=True)
    client_id = sa.Column(sa.String(255), nullable=False)

    receipt_url = sa.Column(sa.String(255))
    request = sa.Column(sa.Text)
    response = sa.Column(sa.Text)
    made_on = sa.Column(sa.DateTime)
    canceled_at = sa.Column(sa.DateTime)
    cancellation_reason = sa.Column(sa.Text)
    payment_method = sa.Column(sa.String(255))

    @declared_attr
    def status_id(self):
        return sa.Column(sa.Integer, sa.ForeignKey(self._status_model.id))

    @declared_attr
    def status(self):
        return sa.orm.relationship(
            self._status_model,
            foreign_keys=self.status_id,
            uselist=False,
            lazy="joined",
        )  # type: ignore


class PaymentRepo:
    def __init__(self, session, model):
        self.model = model
        self.session = session

    def save(self, data: ObjectDict, **kwargs):
        try:
            self.session.add(
                self.model(
                    request_id=data.id,
                    request=json.dumps(data),
                    status_id=PaymentStatusValue.WAIT,
                    **kwargs,
                )
            )
            self.session.commit()
        except sa.exc.SQLAlchemyError:
            self.session.rollback()
            raise

    def get_payment(self, request_id: str) -> Payment:
        return self.model.query.filter_by(request_id=request_id).first_or_404()

    def update(self, status: PaymentStatusValue, event: ObjectDict):
        try:
            data = event.data.object.charges.data[0]
        except (AttributeError, IndexError):
            data = ObjectDict()

        payment = self.get_payment(event.data.object.id)
        payment.status_id = status
        payment.response_id = data.id
        payment.made_on = datetime.now()
        payment.response = json.dumps(data)
        payment.receipt_url = data.receipt_url

        try:
            self.session.add(payment)
            self.session.commit()
        except sa.exc.SQLAlchemyError:
            self.session.rollback()
            raise



from flaskel.views import rpc

from .blueprints import bp_api, bp_auth, bp_spa, bp_web
from .common import (
    ApiDocTemplate,
    ApiSpecTemplate,
    MobileLoggerView,
    MobileReleaseView,
    PasswordForgotView,
    PasswordResetView,
    RegisterView,
    TokenAuthView,
)

rpc_service = object  # import rpc_service here
rpc.JSONRPCView.load_from_object(rpc_service())

BLUEPRINTS = (
    (bp_spa,),
    (bp_api, {"subdomain": bp_api.subdomain}),
    (bp_web, {"url_prefix": bp_web.url_prefix}),
    (bp_auth, {"url_prefix": bp_auth.url_prefix}),
)

VIEWS = (
    (TokenAuthView, bp_auth),
    (RegisterView, bp_auth),
    (PasswordResetView, bp_auth),
    (PasswordForgotView, bp_auth),
    (rpc.JSONRPCView, bp_api),
    (ApiDocTemplate, bp_api),
    (ApiSpecTemplate, bp_api),
    (MobileReleaseView, bp_api),
    (MobileLoggerView, bp_api),
)



from twisted.internet import reactor  # pylint: disable=import-error
from twisted.web.server import Site  # pylint: disable=import-error
from twisted.web.wsgi import WSGIResource  # pylint: disable=import-error

from .base import BaseApplication


class WSGITwisted(BaseApplication):
    def __init__(self, app, options=None):
        BaseApplication.__init__(self, app, options)
        # noinspection PyUnresolvedReferences
        # pylint: disable=no-member
        resource = WSGIResource(reactor, reactor.getThreadPool(), self.application)
        self.site = Site(resource)

    def run(self):
        # noinspection PyUnresolvedReferences
        # pylint: disable=no-member
        reactor.listenTCP(self._port, self.site, interface=self._interface)
        # noinspection PyUnresolvedReferences
        reactor.run()  # pylint: disable=no-member



import os.path
import typing as t

import flask
from vbcore.datastruct import Dumper, ObjectDict
from vbcore.http import httpcode
from vbcore.http.headers import HeaderEnum
from vbcore.json import JsonEncoder
from werkzeug.routing import Rule
from werkzeug.utils import safe_join

cap: "Flaskel" = t.cast("Flaskel", flask.current_app)
request: "Request" = t.cast("Request", flask.request)


class Config(flask.Config, ObjectDict):
    def __init__(self, root_path: str, defaults: t.Optional[dict] = None):
        flask.Config.__init__(self, root_path, defaults)
        ObjectDict.__init__(self)


class DumpUrls(Dumper):
    def __init__(self, app: flask.Flask):
        super().__init__(app.url_map.iter_rules())

    @property
    def rules(self) -> t.Iterator[Rule]:
        return self.data

    def dump(self) -> str:
        output = []
        for rule in self.rules:
            methods = ",".join(rule.methods)
            output.append(f"{rule.endpoint}:30s {methods}:40s {rule}")
        return "\n".join(sorted(output))


class Request(flask.Request):
    @property
    def id(self) -> t.Optional[str]:  # pylint: disable=invalid-name
        hdr = cap.config.REQUEST_ID_HEADER
        if hasattr(flask.g, "request_id"):
            return flask.g.request_id
        if hdr in flask.request.headers:
            return flask.request.headers[hdr]

        if not hdr:
            return None

        flask_header_name = f"HTTP_{hdr.upper().replace('-', '_')}"
        return flask.request.environ.get(flask_header_name)

    def get_json(self, *args, allow_empty=False, **kwargs) -> ObjectDict:
        """

        :param allow_empty:
        :return:
        """
        payload = super().get_json(*args, **kwargs)
        if payload is None:
            if not allow_empty:
                flask.abort(httpcode.BAD_REQUEST, "No JSON in request")
            payload = ObjectDict()  # pragma: no cover

        return ObjectDict.normalize(payload)


class Response(flask.Response):
    @classmethod
    def no_content(cls, status=httpcode.NO_CONTENT, headers=None) -> "Response":
        """

        :param status:
        :param headers:
        :return:
        """
        response = flask.make_response(bytes())
        response.headers.update(headers or {})
        response.headers.pop(HeaderEnum.CONTENT_TYPE)
        response.headers.pop(HeaderEnum.CONTENT_LENGTH)
        response.status_code = status
        return t.cast("Response", response)

    @classmethod
    def set_sendfile_headers(cls, response: "Response", file_path: str) -> "Response":
        hdr = HeaderEnum
        conf = cap.config
        response.headers[hdr.X_ACCEL_REDIRECT] = os.path.abspath(file_path)
        response.headers[hdr.X_ACCEL_CHARSET] = conf.ACCEL_CHARSET or "utf-8"
        response.headers[hdr.X_ACCEL_BUFFERING] = (
            "yes" if conf.ACCEL_BUFFERING else "no"
        )
        if conf.ACCEL_LIMIT_RATE:
            response.headers[hdr.X_ACCEL_LIMIT_RATE] = conf.ACCEL_LIMIT_RATE
        if conf.SEND_FILE_MAX_AGE_DEFAULT:
            response.headers[hdr.X_ACCEL_EXPIRES] = conf.SEND_FILE_MAX_AGE_DEFAULT
        return response

    @classmethod
    def send_file(cls, directory: str, filename: str, **kwargs) -> "Response":
        kwargs.setdefault("as_attachment", True)
        file_path = safe_join(directory, filename)

        try:
            response = flask.send_file(file_path, etag=True, conditional=True, **kwargs)
        except IOError as exc:
            cap.logger.warning(str(exc))
            return flask.abort(httpcode.NOT_FOUND)

        if cap.use_x_sendfile is True and cap.config.ENABLE_ACCEL is True:
            # following headers works with nginx compatible proxy
            return cls.set_sendfile_headers(response, file_path)
        return response

    def get_json(self, *args, **kwargs) -> ObjectDict:
        return ObjectDict.normalize(super().get_json(*args, **kwargs))


class Flaskel(flask.Flask):
    config_class = Config
    request_class = Request
    response_class = Response
    json_encoder = JsonEncoder

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.version = None
        self.config: Config



import typing as t

from tornado.httpserver import HTTPServer  # pylint: disable=import-error
from tornado.ioloop import IOLoop  # pylint: disable=import-error
from tornado.wsgi import WSGIContainer  # pylint: disable=import-error

from flaskel import Flaskel

from .base import BaseApplication


class WSGITornado(BaseApplication):
    def __init__(self, app: Flaskel, options: t.Optional[dict] = None):
        BaseApplication.__init__(self, app, options)
        self.http_server = HTTPServer(WSGIContainer(self.application))

    def run(self):
        self.http_server.listen(address=self._interface, port=self._port)
        IOLoop.current().start()



from vbcore.date_helper import DateHelper


class FlaskDateHelper:
    def __init__(self, app=None, **kwargs):
        self._helper = None
        self.iso_format = None

        if app is not None:
            self.init_app(app, **kwargs)

    def init_app(self, app, helper=DateHelper, **__):
        app.config.setdefault("DATE_HELPER_COUNTRY", "IT")
        app.config.setdefault("DATE_HELPER_PROV", None)
        app.config.setdefault("DATE_HELPER_STATE", None)
        app.config.setdefault("DATE_PRETTY", helper.DATE_PRETTY_FORMAT)
        app.config.setdefault("DATE_ISO_FORMAT", helper.DATE_ISO_FORMAT)

        self._helper = helper
        self.iso_format = app.config.DATE_ISO_FORMAT

        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["date_helper"] = self

    def __getattr__(self, name):
        return getattr(self._helper, name)



import typing as t

import flask
from vbcore.http import HttpMethod

from flaskel.flaskel import Response

from .base import BaseView


class RenderTemplate(BaseView):
    methods: t.List[str] = [
        HttpMethod.GET,
        HttpMethod.POST,
    ]

    template: str = "index.html"

    def __init__(
        self,
        template: t.Optional[str] = None,
        content_type: t.Optional[str] = None,
        **kwargs,
    ):
        self._params = kwargs
        self._content_type = content_type
        self._template = template or self.template

    def service(self, *_, **kwargs):
        return kwargs or self._params

    def response(self, data, **kwargs):
        return Response(data, mimetype=self._content_type, **kwargs)

    def dispatch_request(self, *_, **__):
        params = self.service()
        template = flask.render_template(self._template, **params)
        return self.response(template)


class RenderTemplateString(RenderTemplate):
    def dispatch_request(self, *_, **__):
        params = self.service()
        template = flask.render_template_string(self._template, **params)
        return self.response(template)



import typing as t

from vbcore import uuid
from vbcore.datastruct import ObjectDict
from werkzeug.middleware.proxy_fix import ProxyFix
from werkzeug.urls import url_decode

if t.TYPE_CHECKING:
    # pylint: disable=unused-import
    from _typeshed.wsgi import StartResponse, WSGIApplication, WSGIEnvironment

    from flaskel import Flaskel


class BaseMiddleware:
    def __init__(self, app: "WSGIApplication"):
        self.app = app
        self._flask_app: t.Optional["Flaskel"] = None

    @property
    def flask_app(self) -> t.Optional["Flaskel"]:
        return self._flask_app

    @flask_app.setter
    def flask_app(self, app: "Flaskel"):
        self._flask_app = app

    def get_config(self) -> ObjectDict:
        try:
            if isinstance(self.flask_app.config, ObjectDict):
                return self.flask_app.config
            return ObjectDict.normalize(self.flask_app.config)
        except AttributeError:
            return ObjectDict()


class ForceHttps(BaseMiddleware):
    """
    NOTE: use this only if you can not touch reverse proxy configuration
    use ReverseProxied instead
    """

    def __call__(
        self, environ: "WSGIEnvironment", start_response: "StartResponse"
    ) -> t.Iterable[bytes]:
        environ["wsgi.url_scheme"] = "https"
        return self.app(environ, start_response)


class ReverseProxied(ProxyFix, BaseMiddleware):
    """
    adjust the WSGI environ based on X-Forwarded- that proxies in
    front of the application may set:

    -   X-Forwarded-For    -> sets REMOTE_ADDR
    -   X-Forwarded-Proto  -> sets wsgi.url_scheme
    -   X-Forwarded-Host   -> sets HTTP_HOST, SERVER_NAME, and SERVER_PORT
    -   X-Forwarded-Port   -> sets HTTP_HOST and SERVER_PORT
    -   X-Forwarded-Prefix -> sets SCRIPT_NAME

    for example nginx:
        location /prefix {
            proxy_pass http://127.0.0.1:5000;
            proxy_set_header Host $host;
            proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
            proxy_set_header X-Forwarded-Proto $proto;
            proxy_set_header X-Forwarded-Port 5000;
            proxy_set_header X-Forwarded-Prefix /prefix;
        }

    You must tell the middleware how many proxies set each header so it
    knows what values to trust. It is a security issue to trust values
    that came from the client rather than a proxy.

    The original values of the headers are stored in the WSGI
    environ as werkzeug.proxy_fix.orig, a dict.
    """

    def __init__(
        self,
        app: "WSGIApplication",
        x_for: int = 1,
        x_proto: int = 1,
        x_host: int = 1,
        x_port: int = 0,
        x_prefix: int = 1,
    ):
        ProxyFix.__init__(self, app, x_for, x_proto, x_host, x_port, x_prefix)
        BaseMiddleware.__init__(self, app)


class HTTPMethodOverride(BaseMiddleware):
    """
    Implements the hidden HTTP method technique.
    Not all web browsers or reverse proxy supports every HTTP method.
    Client can use 'X-HTTP-Method-Override' header or '_method_override' in querystring.
    Only POST method can be overridden because with GET requests
    could result unexpected behavior due to caching.
    """

    def __call__(
        self, environ: "WSGIEnvironment", start_response: "StartResponse"
    ) -> t.Iterable[bytes]:
        conf = self.get_config()
        methods = conf.OVERRIDE_METHODS or ("POST",)
        method = environ.get("HTTP_X_HTTP_METHOD_OVERRIDE")

        if "_method_override" in environ.get("QUERY_STRING", ""):
            args = url_decode(environ["QUERY_STRING"])
            method = args.get("_method_override")

        if method and environ["REQUEST_METHOD"] in methods:
            environ["REQUEST_METHOD"] = method.upper()

        return self.app(environ, start_response)


class RequestID(BaseMiddleware):
    """
    From: https://github.com/antarctica/flask-request-id-header

    Flask middleware to ensure all requests have a Request ID header present.
    Request IDs can be used when logging errors and allows users to trace requests through multiple
    layers such as load balancers and reverse proxies.
    The request ID header value may consist of multiple IDs encoded according to RFC 2616 [1].

    This middleware ensures there is always at least one, unique, value,
    whilst respecting any pre-existing values set by an upstream component or the client.
    If there isn't a pre-existing value, a single, unique, ID is generated.
    If there is a pre-existing value, it is split into multiple values as per RFC 2616
    and each value checked for uniqueness.
    If no values are unique an additional, unique, value is added.

    A value is considered unique if it:
    1. is a valid UUID (version 4)
    2. or, the value is known to be from a component that assigns unique IDs,
       identified by a common prefix set in the 'REQUEST_ID_PREFIX' Flask config value
       or `request_id_prefix` class member value.

    To use the Request ID in the current application: `request.environ.get("HTTP_X_REQUEST_ID")`
    For use elsewhere, the Request ID header is included in the response back to the client.
    [1] http://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.2
    """

    # the http header name, overridden via flask app config
    header_name = "X-Request-ID"

    # the request id value prefix used to trust uniqueness
    request_id_prefix = None

    def __call__(
        self, environ: "WSGIEnvironment", start_response: "StartResponse"
    ) -> t.Iterable[bytes]:
        conf = self.get_config()
        header_name = conf.REQUEST_ID_HEADER or self.header_name
        flask_header_name = f"HTTP_{header_name.upper().replace('-', '_')}"
        request_id_header = self.prepare(environ.get(flask_header_name))
        environ[flask_header_name] = request_id_header

        def new_start_response(status, headers, exc_info=None):
            headers.append((header_name, request_id_header))
            return start_response(status, headers, exc_info)

        return self.app(environ, new_start_response)

    def prepare(self, header_value: t.Optional[str] = None) -> str:
        """
        Ensures request ID header has at least one, unique, value

        :param header_value: Existing request ID HTTP header value
        :return: Computed Request ID HTTP header
        """
        if header_value is None:
            return uuid.get_uuid()

        for request_id in header_value.split(","):
            if self.has_valid_uniqueness(request_id):
                return header_value

        # Append an unique header value
        return f"{header_value},{uuid.get_uuid()}"

    def has_valid_uniqueness(self, request_id: str) -> bool:
        """
        Checks whether a Request ID has a trusted prefix or is a valid uuid

        :param request_id: A request ID
        :return: Whether the Request ID is unique or not
        """
        conf = self.get_config()
        prefix = conf.REQUEST_ID_PREFIX or self.request_id_prefix
        if prefix and request_id.startswith(prefix):
            return True  # pragma: no cover

        return uuid.check_uuid(request_id, exc=False)



from flask import url_for
from vbcore.datastruct import ObjectDict
from vbcore.http import httpcode

from flaskel.flaskel import cap, request
from flaskel.http.exceptions import abort
from flaskel.views.base import BaseView
from flaskel.views.template import RenderTemplateString


class ApiDocTemplate(RenderTemplateString):
    apispec_view = "api.apispec"
    default_view_name = "apidocs"
    default_urls = ("/apidocs",)

    template = """<!doctype html>
<html>
    <head>
        <meta charset="utf-8"/>
        <meta http-equiv="X-UA-Compatible" content="IE=edge,chrome=1"/>
        <meta name="viewport" content="width=device-width,initial-scale=1,shrink-to-fit=no"/>
        <title>{{page_title}}</title>
        <script type="module"
            src="https://unpkg.com/rapidoc@{{rapidoc_version}}/dist/rapidoc-min.js">
        </script>
    </head>
    <body>
        <rapi-doc
            theme="{{rapidoc_theme}}" spec-url="{{spec_url}}"
            render-style="view" font-size="large"
        ></rapi-doc>
    </body>
</html>"""

    def service(self, *_, **__) -> dict:
        if not cap.config.APIDOCS_ENABLED:
            return abort(httpcode.NOT_FOUND)

        proto = cap.config.PREFERRED_URL_SCHEME
        return dict(
            rapidoc_version=cap.config.RAPIDOC_VERSION or "9",
            rapidoc_theme=cap.config.RAPIDOC_THEME or "dark",
            page_title=f"{cap.config.APP_NAME} - API DOCS",
            spec_url=url_for(self.apispec_view, _external=True, _scheme=proto),
        )


class ApiSpecTemplate(BaseView):
    default_view_name = "apispec"
    default_urls = ("/apidoc.json",)

    def __init__(self, version: str = "1.0.0", context_path: str = ""):
        self._api_version = version
        self._context_path = f"{request.environ['SCRIPT_NAME']}/{context_path}"

    def dispatch_request(self, *_, **__):
        if not cap.config.APIDOCS_ENABLED:
            abort(httpcode.NOT_FOUND)

        return self.set_variables(cap.config.APISPEC)

    def set_variables(self, apispec: ObjectDict) -> ObjectDict:
        try:
            apispec.info.version = self._api_version
            variables = apispec.servers[0].variables
            variables["context"]["default"] = self._context_path
            scheme = cap.config.PREFERRED_URL_SCHEME or "http"
            variables["host"]["default"] = f"{scheme}://{cap.config.SERVER_NAME}"
        except Exception as exc:  # pylint: disable=broad-except
            cap.logger.exception(exc)

        return apispec



from vbcore.http import httpcode
from vbcore.http.headers import ContentTypeEnum, HeaderEnum
from vbcore.tester.helpers import basic_auth_header
from vbcore.tester.mixins import Asserter

from flaskel.tester.helpers import url_for


def test_app_runs(api_tester):
    api_tester.get(
        url="/",
        status=httpcode.NOT_FOUND,
        mimetype=ContentTypeEnum.JSON_PROBLEM,
    )


def test_cors_headers(api_tester, config):
    res = api_tester.get(
        url="/", status=httpcode.NOT_FOUND, mimetype=ContentTypeEnum.JSON_PROBLEM
    )
    Asserter.assert_header(res, HeaderEnum.ACCESS_CONTROL_ALLOW_ORIGIN, "*")
    Asserter.assert_allin(
        res.headers[HeaderEnum.ACCESS_CONTROL_EXPOSE_HEADERS].split(", "),
        config.CORS_EXPOSE_HEADERS,
    )


def test_apidoc(config, api_tester, views):
    headers = basic_auth_header(config.ADMIN_EMAIL, config.ADMIN_PASSWORD)

    api_tester.get(
        url=url_for(views.api_docs), headers=headers, mimetype=ContentTypeEnum.HTML
    )
    res = api_tester.get(
        url=url_for(views.api_spec), headers=headers, mimetype=ContentTypeEnum.JSON
    )
    Asserter.assert_different(res.json, {})


def test_jwt(api_tester, config, auth_token, views):
    tokens = api_tester.post(
        url_for(views.access_token),
        json=dict(email=config.ADMIN_EMAIL, password=config.ADMIN_PASSWORD),
        schema=config.SCHEMAS.AccessToken,
    )
    token_refresh = auth_token(tokens.json.refresh_token)

    api_tester.post(
        url=url_for(views.access_token),
        json=dict(email=config.ADMIN_EMAIL, password="bad password"),
        status=httpcode.UNAUTHORIZED,
    )

    api_tester.post(
        url=url_for(views.refresh_token),
        headers=token_refresh,
        schema=config.SCHEMAS.RefreshToken,
    )

    api_tester.post(
        url=url_for(views.revoke_token),
        json=dict(
            access_token=tokens.json.access_token,
            refresh_token=tokens.json.refresh_token,
        ),
        status=httpcode.NO_CONTENT,
    )

    api_tester.post(
        url=url_for(views.refresh_token),
        headers=token_refresh,
        status=httpcode.UNAUTHORIZED,
    )



from models.user import AccountModel
from vbcore.datastruct import ObjectDict

from flaskel import client_redis
from flaskel.ext import auth
from flaskel.extra.account import AccountHandler


class TokenHandler(auth.RedisTokenHandler):
    def prepare_identity(self, data, **__):
        return ObjectDict(id=data.id, email=data.email)


token_handler = TokenHandler(client_redis)
account_handler = AccountHandler(model=AccountModel)



import enum
import typing as t
from functools import partial, wraps

from flask import Blueprint
from vbcore.batch import BatchExecutor
from vbcore.datastruct import LStrEnum
from vbcore.http import httpcode, HttpMethod
from vbcore.http.headers import ContentTypeEnum, HeaderEnum

from flaskel.flaskel import Flaskel, request

from .checkers import CheckerResponseType

DecoratorType = t.Callable[[t.Callable], None]
ExecutorType = t.TypeVar("ExecutorType", bound=BatchExecutor)
CheckerType = t.Tuple[t.Callable, t.Optional[t.Dict[str, t.Any]]]


class CheckStatus(LStrEnum):
    FAIL = enum.auto()
    PASS = enum.auto()
    WARN = enum.auto()


class HealthCheck:
    def __init__(self, app: t.Optional[Flaskel] = None, **kwargs):
        self._executor = BatchExecutor
        self._checkers: t.Dict[str, t.Callable] = {}
        self.app: t.Optional[Flaskel] = t.cast(Flaskel, None)

        if app is not None:
            self.init_app(app, **kwargs)

    def init_app(
        self,
        app: Flaskel,
        blueprint: t.Optional[Blueprint] = None,
        checkers: t.Tuple[CheckerType, ...] = (),
        decorators: t.Tuple[DecoratorType, ...] = (),
        executor: t.Optional[ExecutorType] = None,
    ):
        """

        :param app: Flaskel instance
        :param blueprint: optional a blueprint instance on which register the healthcheck view
        :param checkers: functions with arguments that execute the checks
        :param decorators: functions that decorates the healthcheck view
        :param executor: batch executor class, must be subclass of BatchExecutor
        """
        self.app = app
        self._executor = executor or self._executor

        self.set_default_config()
        self.register_checkers(checkers)
        self.register_route(blueprint or app, decorators)

        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["healthcheck"] = self

    def set_default_config(self):
        self.app.config.setdefault("HEALTHCHECK_ABOUT_LINK", None)
        self.app.config.setdefault("HEALTHCHECK_VIEW_NAME", "healthcheck")
        self.app.config.setdefault("HEALTHCHECK_PATH", "/healthcheck")
        self.app.config.setdefault("HEALTHCHECK_PARAM_KEY", "checks")
        self.app.config.setdefault("HEALTHCHECK_PARAM_SEP", "+")
        self.app.config.setdefault(
            "HEALTHCHECK_CONTENT_TYPE", ContentTypeEnum.JSON_HEALTH
        )

    def register_route(
        self,
        blueprint: t.Optional[t.Union[Flaskel, Blueprint]] = None,
        decorators: t.Tuple[DecoratorType, ...] = (),
    ) -> t.Callable:
        view = self.perform
        for decorator in decorators:
            view = decorator(view)

        blueprint.add_url_rule(
            self.app.config.HEALTHCHECK_PATH,
            self.app.config.HEALTHCHECK_VIEW_NAME,
            methods=[HttpMethod.GET],
            view_func=view,
        )
        return view

    def register_checkers(self, checkers: t.Tuple[CheckerType, ...]):
        for func, kwargs in checkers:
            self.register(**(kwargs or {}))(func)

    def get_checkers(self) -> t.List[str]:
        params = request.args.get(self.app.config.HEALTHCHECK_PARAM_KEY)
        only_checkers = (
            params.split(self.app.config.HEALTHCHECK_PARAM_SEP) if params else None
        )
        all_checkers = set(self._checkers)
        return list(
            all_checkers
            if not only_checkers
            else set(only_checkers).intersection(all_checkers)
        )

    def execute(self) -> t.Tuple[t.List[str], t.List[CheckerResponseType]]:
        checkers = self.get_checkers()
        tasks = [(self._checkers.get(c), dict(app=self.app)) for c in checkers]
        return checkers, self._executor(tasks=tasks).run()

    def perform(self) -> t.Tuple[dict, int, dict]:
        healthy = True
        response: t.Dict[str, t.Any] = {
            "status": CheckStatus.PASS,
            "checks": {},
            "links": {"about": self.app.config.HEALTHCHECK_ABOUT_LINK},
        }

        checkers, checkers_responses = self.execute()
        for index, name in enumerate(checkers):
            state, message = checkers_responses[index]
            status = CheckStatus.PASS if state else CheckStatus.FAIL
            response["checks"][name] = {"status": status, "output": message}
            if not state:
                healthy = False

        if not healthy:
            response["status"] = CheckStatus.FAIL

        return (
            response,
            httpcode.SUCCESS if healthy else httpcode.MULTI_STATUS,
            {HeaderEnum.CONTENT_TYPE: self.app.config.HEALTHCHECK_CONTENT_TYPE},
        )

    def register(self, name: t.Optional[str] = None, **kwargs):
        def decorator(func):
            @wraps(func)
            def wrapped():
                self._checkers[name or func.__name__] = partial(func, **kwargs)

            return wrapped()

        return decorator



import flask

# Define your forms in this package and import them here


def flash_errors(form, redirect=None):
    for field, errors in form.errors.items():
        for e in errors:
            flask.flash(f"{getattr(form, field).label.text}: {e}")
        if redirect:
            return flask.redirect(flask.url_for(redirect))



import functools
import typing as t

from flask_cloudflare_remote import CloudflareRemote
from flask_cors import CORS
from flask_errors_handler import ErrorHandler
from flask_logify import FlaskLogging
from flask_response_builder import ResponseBuilder
from flask_sqlalchemy import SQLAlchemy
from flask_template_support import TemplateSupport

from .caching import Caching
from .crypto.argon import Argon2
from .datetime import FlaskDateHelper
from .errors import ErrorNormalizer
from .healthcheck import HealthCheck
from .jobs import APJobs
from .sendmail import ClientMail
from .sqlalchemy import SQLAModel
from .useragent import UserAgent

cors: CORS = CORS()
logger: FlaskLogging = FlaskLogging()
builder: ResponseBuilder = ResponseBuilder()
template: TemplateSupport = TemplateSupport()
cfremote: CloudflareRemote = CloudflareRemote()
error_handler: ErrorHandler = ErrorHandler(normalizer=ErrorNormalizer())

argon2: Argon2 = Argon2()
caching: Caching = Caching()
useragent: UserAgent = UserAgent()
client_mail: ClientMail = ClientMail()
health_checks: HealthCheck = HealthCheck()
date_helper: FlaskDateHelper = FlaskDateHelper()

Scheduler: t.Type[APJobs] = t.cast(
    t.Type[APJobs],
    functools.partial(APJobs),
)

Database: t.Type[SQLAlchemy] = t.cast(
    t.Type[SQLAlchemy],
    functools.partial(SQLAlchemy, model_class=SQLAModel),
)



import typing as t

import jsonschema
from vbcore.datastruct import ObjectDict
from vbcore.http import httpcode
from vbcore.jsonschema.support import JSONSchema

from flaskel import cap, request
from flaskel.http.exceptions import abort

from .datastruct import ConfigProxy


class PayloadValidator:
    schemas: ObjectDict = ConfigProxy("SCHEMAS")
    validator: t.Type[JSONSchema] = JSONSchema

    @classmethod
    def validate(cls, schema: t.Union[str, dict], strict: bool = True) -> ObjectDict:
        if strict and schema is None:
            abort(httpcode.INTERNAL_SERVER_ERROR, "empty schema")

        payload = request.json
        try:
            schema = cls.schemas.get(schema) if isinstance(schema, str) else schema
            cls.validator.validate(payload, schema, raise_exc=True)
            return payload
        except jsonschema.SchemaError as exc:
            cap.logger.exception(exc)
            return abort(httpcode.INTERNAL_SERVER_ERROR)
        except jsonschema.ValidationError as exc:
            cap.logger.error(cls.validator.error_report(exc, payload))
            reason = dict(cause=exc.cause, message=exc.message, path=exc.path)
            return abort(httpcode.UNPROCESSABLE_ENTITY, response=dict(reason=reason))



__version__ = "3.0.0.rc3"

__author_name__ = "cs91chris"
__author_email__ = "cs91chris@voidbrain.me"

__author__ = f"{__author_name__} <{__author_email__}>"

__all__ = [
    "__version__",
    "__author__",
    "__author_name__",
    "__author_email__",
]



from .checkers import (
    CheckerResponseType,
    health_mongo,
    health_redis,
    health_services,
    health_sqlalchemy,
    health_system,
    SuccessResponse,
)
from .health import HealthCheck



import os
import typing as t

from vbcore.uuid import get_uuid
from werkzeug.datastructures import FileStorage
from werkzeug.utils import secure_filename

from flaskel import cap, ConfigProxy

from .exceptions import MediaError
from .repo import MediaRepo


class MediaService:
    media_repo = MediaRepo
    obfuscate_filename = True
    config = ConfigProxy("MEDIA")

    @classmethod
    def get_ext(cls, filename) -> t.Optional[str]:
        if "." in filename:
            return filename.split(".", 1)[-1].lower()
        return None

    @classmethod
    def generate_filepath(cls, eid, filename: str) -> t.Tuple[str, str]:
        ext = cls.get_ext(filename)
        entity_name = cls.media_repo.entity_name()
        if cls.obfuscate_filename is True:
            filename = get_uuid()
        else:
            filename = secure_filename(filename)

        filename = f"{eid}-{filename}.{ext}"
        filepath = os.path.join(entity_name, filename)
        url = "/".join((entity_name, filename))
        return url, filepath

    @classmethod
    def check_file(cls, file: FileStorage):
        if not file.filename:
            raise MediaError("missing file name")
        if cls.get_ext(file.filename) not in cls.config.ALLOWED_EXTENSIONS:
            raise MediaError("file extension not allowed")

    @classmethod
    def upload(cls, files: t.Iterable[FileStorage], eid) -> list:
        medias = []
        emodel = cls.media_repo.get(eid)
        for f in files:
            cls.check_file(f)
            url, filepath = cls.generate_filepath(emodel.id, f.filename)
            f.save(os.path.join(cls.config.UPLOAD_FOLDER, filepath))
            res = cls.media_repo.update(
                emodel, link=url, path=filepath, filename=f.filename
            )
            medias.append(res)

        cls.media_repo.store(emodel)
        return medias

    @classmethod
    def delete(cls, eid, res_id):
        res = cls.media_repo.delete(eid, res_id)

        try:
            filepath = os.path.join(*cls.config.UPLOAD_FOLDER.split("/"), res.path)
            if os.path.isfile(filepath):
                os.remove(filepath)
        except OSError as exc:
            cap.logger.exception(exc)



import os
import shutil
import sys
from pathlib import Path


def replace_in_file(file, *args):
    _f = Path(file)
    text = _f.read_text(encoding="utf-8")
    for sd in args:
        text = text.replace(*sd)
    _f.write_text(text, encoding="utf-8")


def copy_skeleton(name: str):
    from flaskel import scripts as flaskel_scripts  # pylint: disable=C0415

    try:
        source = os.path.join(flaskel_scripts.__path__[0], "skeleton")
        shutil.copytree(source, ".", dirs_exist_ok=True)
        shutil.move("skel", name)
    except OSError as e:
        print(f"Unable to create new app. Error: {e}", file=sys.stderr)
        sys.exit(1)


def fix_files_references(name: str, placeholder: str):
    for args in (
        (
            os.path.join(name, "scripts", "cli.py"),
            ("from ext", f"from {name}.ext"),
            ("from views", f"from {name}.views"),
        ),
        (
            os.path.join("tests", "conftest.py"),
            ("from scripts.cli", f"from {name}.scripts.cli"),
        ),
        (
            os.path.join("tests", "conftest.py"),
            (placeholder, name),
        ),
    ):
        replace_in_file(*args)


def init_app(name: str, placeholder: str = "{skeleton}"):
    copy_skeleton(name)
    init_file = Path(os.path.join(name, "__init__.py"))
    init_file.write_text("from .version import *\n", encoding="utf-8")
    fix_files_references(name, placeholder)



from multiprocessing import cpu_count

from waitress import serve  # pylint: disable=import-error

from .base import BaseApplication


class WSGIWaitress(BaseApplication):
    def run(self):
        self.options.setdefault("threads", cpu_count())

        serve(
            self.application,
            host=self._interface,
            port=self._port,
            threads=self.options["threads"],
        )



from flaskel import client_redis
from flaskel.ext.auth import basic_auth, RedisTokenHandler
from flaskel.ext.default import caching
from flaskel.ext.limit import RateLimit
from flaskel.extra import apidoc, mobile_support as mobile
from flaskel.extra.account import (
    PasswordForgotView as BasePasswordForgotView,
    PasswordResetView as BasePasswordResetView,
    RegisterView as BaseRegisterView,
)
from flaskel.views import Resource
from flaskel.views.token import BaseTokenAuth

from ..ext.auth import token_handler


class TokenAuthView(BaseTokenAuth):
    handler = RedisTokenHandler(client_redis)

    @classmethod
    def check_credential(cls):
        raise NotImplementedError("implement me!")


class ApiDocTemplate(apidoc.ApiDocTemplate):
    decorators = [
        caching.cached(),
        basic_auth.login_required(),
    ]


class ApiSpecTemplate(apidoc.ApiSpecTemplate):
    decorators = [
        caching.cached(),
        basic_auth.login_required(),
    ]


class MobileReleaseView(mobile.MobileReleaseView):
    decorators = [
        token_handler.auth_required(),
    ]


class MobileLoggerView(mobile.MobileLoggerView):
    decorators = [
        token_handler.auth_required(),
    ]


class RegisterView(BaseRegisterView):
    decorators = [
        RateLimit.slow(),
    ]


class PasswordForgotView(BasePasswordForgotView):
    decorators = [
        RateLimit.slow(),
    ]


class PasswordResetView(BasePasswordResetView):
    decorators = [
        token_handler.auth_required(),
    ]


class ApiCatalog(Resource):
    methods_subresource = None

    decorators = [
        caching.cached(),
        token_handler.auth_required(),
    ]



REGEX_OBJECT_KEY = "^[a-zA-Z0-9\\.\\-_]+$"


class Definitions:
    reference = "#/definitions/Reference"
    schema = "#/definitions/Schema"
    server = "#/definitions/Server"
    external_documentation = "#/definitions/ExternalDocumentation"
    response = "#/definitions/Response"
    header = "#/definitions/Header"
    parameter = "#/definitions/Parameter"
    example = "#/definitions/Example"
    mediatype = "#/definitions/MediaType"
    example_xor_examples = "#/definitions/ExampleXORExamples"


# pylint: disable=C0302
SCHEMA = {
    "id": "https://spec.openapis.org/oas/3.0/schema/2019-04-02",
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Validation schema for OpenAPI Specification 3.0.X.",
    "type": "object",
    "required": ["openapi", "info", "paths"],
    "properties": {
        "openapi": {"type": "string", "pattern": "^3\\.0\\.\\d(-.+)?$"},
        "info": {"$ref": "#/definitions/Info"},
        "externalDocs": {"$ref": Definitions.external_documentation},
        "servers": {"type": "array", "items": {"$ref": Definitions.server}},
        "security": {
            "type": "array",
            "items": {"$ref": "#/definitions/SecurityRequirement"},
        },
        "tags": {
            "type": "array",
            "items": {"$ref": "#/definitions/Tag"},
            "uniqueItems": True,
        },
        "paths": {"$ref": "#/definitions/Paths"},
        "components": {"$ref": "#/definitions/Components"},
    },
    "patternProperties": {"^x-": {}},
    "additionalProperties": False,
    "definitions": {
        "Reference": {
            "type": "object",
            "required": ["$ref"],
            "patternProperties": {
                "^\\$ref$": {"type": "string", "format": "uri-reference"}
            },
        },
        "Info": {
            "type": "object",
            "required": ["title", "version"],
            "properties": {
                "title": {"type": "string"},
                "description": {"type": "string"},
                "termsOfService": {"type": "string", "format": "uri-reference"},
                "contact": {"$ref": "#/definitions/Contact"},
                "license": {"$ref": "#/definitions/License"},
                "version": {"type": "string"},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "Contact": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "url": {"type": "string", "format": "uri-reference"},
                "email": {"type": "string", "format": "email"},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "License": {
            "type": "object",
            "required": ["name"],
            "properties": {
                "name": {"type": "string"},
                "url": {"type": "string", "format": "uri-reference"},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "Server": {
            "type": "object",
            "required": ["url"],
            "properties": {
                "url": {"type": "string"},
                "description": {"type": "string"},
                "variables": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/definitions/ServerVariable"},
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "ServerVariable": {
            "type": "object",
            "required": ["default"],
            "properties": {
                "enum": {"type": "array", "items": {"type": "string"}},
                "default": {"type": "string"},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "Components": {
            "type": "object",
            "properties": {
                "schemas": {
                    "type": "object",
                    "patternProperties": {
                        REGEX_OBJECT_KEY: {
                            "oneOf": [
                                {"$ref": Definitions.schema},
                                {"$ref": Definitions.reference},
                            ]
                        }
                    },
                },
                "responses": {
                    "type": "object",
                    "patternProperties": {
                        REGEX_OBJECT_KEY: {
                            "oneOf": [
                                {"$ref": Definitions.reference},
                                {"$ref": Definitions.response},
                            ]
                        }
                    },
                },
                "parameters": {
                    "type": "object",
                    "patternProperties": {
                        REGEX_OBJECT_KEY: {
                            "oneOf": [
                                {"$ref": Definitions.reference},
                                {"$ref": Definitions.parameter},
                            ]
                        }
                    },
                },
                "examples": {
                    "type": "object",
                    "patternProperties": {
                        REGEX_OBJECT_KEY: {
                            "oneOf": [
                                {"$ref": Definitions.reference},
                                {"$ref": Definitions.example},
                            ]
                        }
                    },
                },
                "requestBodies": {
                    "type": "object",
                    "patternProperties": {
                        REGEX_OBJECT_KEY: {
                            "oneOf": [
                                {"$ref": Definitions.reference},
                                {"$ref": "#/definitions/RequestBody"},
                            ]
                        }
                    },
                },
                "headers": {
                    "type": "object",
                    "patternProperties": {
                        REGEX_OBJECT_KEY: {
                            "oneOf": [
                                {"$ref": Definitions.reference},
                                {"$ref": Definitions.header},
                            ]
                        }
                    },
                },
                "securitySchemes": {
                    "type": "object",
                    "patternProperties": {
                        REGEX_OBJECT_KEY: {
                            "oneOf": [
                                {"$ref": Definitions.reference},
                                {"$ref": "#/definitions/SecurityScheme"},
                            ]
                        }
                    },
                },
                "links": {
                    "type": "object",
                    "patternProperties": {
                        REGEX_OBJECT_KEY: {
                            "oneOf": [
                                {"$ref": Definitions.reference},
                                {"$ref": "#/definitions/Link"},
                            ]
                        }
                    },
                },
                "callbacks": {
                    "type": "object",
                    "patternProperties": {
                        REGEX_OBJECT_KEY: {
                            "oneOf": [
                                {"$ref": Definitions.reference},
                                {"$ref": "#/definitions/Callback"},
                            ]
                        }
                    },
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "Schema": {
            "type": "object",
            "properties": {
                "title": {"type": "string"},
                "multipleOf": {
                    "type": "number",
                    "minimum": 0,
                    "exclusiveMinimum": True,
                },
                "maximum": {"type": "number"},
                "exclusiveMaximum": {"type": "boolean", "default": False},
                "minimum": {"type": "number"},
                "exclusiveMinimum": {"type": "boolean", "default": False},
                "maxLength": {"type": "integer", "minimum": 0},
                "minLength": {"type": "integer", "minimum": 0, "default": 0},
                "pattern": {"type": "string", "format": "regex"},
                "maxItems": {"type": "integer", "minimum": 0},
                "minItems": {"type": "integer", "minimum": 0, "default": 0},
                "uniqueItems": {"type": "boolean", "default": False},
                "maxProperties": {"type": "integer", "minimum": 0},
                "minProperties": {"type": "integer", "minimum": 0, "default": 0},
                "required": {
                    "type": "array",
                    "items": {"type": "string"},
                    "minItems": 1,
                    "uniqueItems": True,
                },
                "enum": {
                    "type": "array",
                    "items": {},
                    "minItems": 1,
                    "uniqueItems": False,
                },
                "type": {
                    "type": "string",
                    "enum": [
                        "array",
                        "boolean",
                        "integer",
                        "number",
                        "object",
                        "string",
                    ],
                },
                "not": {
                    "oneOf": [
                        {"$ref": Definitions.schema},
                        {"$ref": Definitions.reference},
                    ]
                },
                "allOf": {
                    "type": "array",
                    "items": {
                        "oneOf": [
                            {"$ref": Definitions.schema},
                            {"$ref": Definitions.reference},
                        ]
                    },
                },
                "oneOf": {
                    "type": "array",
                    "items": {
                        "oneOf": [
                            {"$ref": Definitions.schema},
                            {"$ref": Definitions.reference},
                        ]
                    },
                },
                "anyOf": {
                    "type": "array",
                    "items": {
                        "oneOf": [
                            {"$ref": Definitions.schema},
                            {"$ref": Definitions.reference},
                        ]
                    },
                },
                "items": {
                    "oneOf": [
                        {"$ref": Definitions.schema},
                        {"$ref": Definitions.reference},
                    ]
                },
                "properties": {
                    "type": "object",
                    "additionalProperties": {
                        "oneOf": [
                            {"$ref": Definitions.schema},
                            {"$ref": Definitions.reference},
                        ]
                    },
                },
                "additionalProperties": {
                    "oneOf": [
                        {"$ref": Definitions.schema},
                        {"$ref": Definitions.reference},
                        {"type": "boolean"},
                    ],
                    "default": True,
                },
                "description": {"type": "string"},
                "format": {"type": "string"},
                "default": {},
                "nullable": {"type": "boolean", "default": False},
                "discriminator": {"$ref": "#/definitions/Discriminator"},
                "readOnly": {"type": "boolean", "default": False},
                "writeOnly": {"type": "boolean", "default": False},
                "example": {},
                "externalDocs": {"$ref": Definitions.external_documentation},
                "deprecated": {"type": "boolean", "default": False},
                "xml": {"$ref": "#/definitions/XML"},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "Discriminator": {
            "type": "object",
            "required": ["propertyName"],
            "properties": {
                "propertyName": {"type": "string"},
                "mapping": {
                    "type": "object",
                    "additionalProperties": {"type": "string"},
                },
            },
        },
        "XML": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "namespace": {"type": "string", "format": "uri"},
                "prefix": {"type": "string"},
                "attribute": {"type": "boolean", "default": False},
                "wrapped": {"type": "boolean", "default": False},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "Response": {
            "type": "object",
            "required": ["description"],
            "properties": {
                "description": {"type": "string"},
                "headers": {
                    "type": "object",
                    "additionalProperties": {
                        "oneOf": [
                            {"$ref": Definitions.header},
                            {"$ref": Definitions.reference},
                        ]
                    },
                },
                "content": {
                    "type": "object",
                    "additionalProperties": {"$ref": Definitions.mediatype},
                },
                "links": {
                    "type": "object",
                    "additionalProperties": {
                        "oneOf": [
                            {"$ref": "#/definitions/Link"},
                            {"$ref": Definitions.reference},
                        ]
                    },
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "MediaType": {
            "type": "object",
            "properties": {
                "schema": {
                    "oneOf": [
                        {"$ref": Definitions.schema},
                        {"$ref": Definitions.reference},
                    ]
                },
                "example": {},
                "examples": {
                    "type": "object",
                    "additionalProperties": {
                        "oneOf": [
                            {"$ref": Definitions.example},
                            {"$ref": Definitions.reference},
                        ]
                    },
                },
                "encoding": {
                    "type": "object",
                    "additionalProperties": {"$ref": "#/definitions/Encoding"},
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
            "allOf": [{"$ref": Definitions.example_xor_examples}],
        },
        "Example": {
            "type": "object",
            "properties": {
                "summary": {"type": "string"},
                "description": {"type": "string"},
                "value": {},
                "externalValue": {"type": "string", "format": "uri-reference"},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "Header": {
            "type": "object",
            "properties": {
                "description": {"type": "string"},
                "required": {"type": "boolean", "default": False},
                "deprecated": {"type": "boolean", "default": False},
                "allowEmptyValue": {"type": "boolean", "default": False},
                "style": {"type": "string", "enum": ["simple"], "default": "simple"},
                "explode": {"type": "boolean"},
                "allowReserved": {"type": "boolean", "default": False},
                "schema": {
                    "oneOf": [
                        {"$ref": Definitions.schema},
                        {"$ref": Definitions.reference},
                    ]
                },
                "content": {
                    "type": "object",
                    "additionalProperties": {"$ref": Definitions.mediatype},
                    "minProperties": 1,
                    "maxProperties": 1,
                },
                "example": {},
                "examples": {
                    "type": "object",
                    "additionalProperties": {
                        "oneOf": [
                            {"$ref": Definitions.example},
                            {"$ref": Definitions.reference},
                        ]
                    },
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
            "allOf": [
                {"$ref": Definitions.example_xor_examples},
                {"$ref": "#/definitions/SchemaXORContent"},
            ],
        },
        "Paths": {
            "type": "object",
            "patternProperties": {
                "^\\/": {"$ref": "#/definitions/PathItem"},
                "^x-": {},
            },
            "additionalProperties": False,
        },
        "PathItem": {
            "type": "object",
            "properties": {
                "$ref": {"type": "string"},
                "summary": {"type": "string"},
                "description": {"type": "string"},
                "servers": {"type": "array", "items": {"$ref": Definitions.server}},
                "parameters": {
                    "type": "array",
                    "items": {
                        "oneOf": [
                            {"$ref": Definitions.parameter},
                            {"$ref": Definitions.reference},
                        ]
                    },
                    "uniqueItems": True,
                },
            },
            "patternProperties": {
                "^(get|put|post|delete|options|head|patch|trace)$": {
                    "$ref": "#/definitions/Operation"
                },
                "^x-": {},
            },
            "additionalProperties": False,
        },
        "Operation": {
            "type": "object",
            "required": ["responses"],
            "properties": {
                "tags": {"type": "array", "items": {"type": "string"}},
                "summary": {"type": "string"},
                "description": {"type": "string"},
                "externalDocs": {"$ref": Definitions.external_documentation},
                "operationId": {"type": "string"},
                "parameters": {
                    "type": "array",
                    "items": {
                        "oneOf": [
                            {"$ref": Definitions.parameter},
                            {"$ref": Definitions.reference},
                        ]
                    },
                    "uniqueItems": True,
                },
                "requestBody": {
                    "oneOf": [
                        {"$ref": "#/definitions/RequestBody"},
                        {"$ref": Definitions.reference},
                    ]
                },
                "responses": {"$ref": "#/definitions/Responses"},
                "callbacks": {
                    "type": "object",
                    "additionalProperties": {
                        "oneOf": [
                            {"$ref": "#/definitions/Callback"},
                            {"$ref": Definitions.reference},
                        ]
                    },
                },
                "deprecated": {"type": "boolean", "default": False},
                "security": {
                    "type": "array",
                    "items": {"$ref": "#/definitions/SecurityRequirement"},
                },
                "servers": {"type": "array", "items": {"$ref": Definitions.server}},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "Responses": {
            "type": "object",
            "properties": {
                "default": {
                    "oneOf": [
                        {"$ref": Definitions.response},
                        {"$ref": Definitions.reference},
                    ]
                }
            },
            "patternProperties": {
                "^[1-5](?:\\d{2}|XX)$": {
                    "oneOf": [
                        {"$ref": Definitions.response},
                        {"$ref": Definitions.reference},
                    ]
                },
                "^x-": {},
            },
            "minProperties": 1,
            "additionalProperties": False,
        },
        "SecurityRequirement": {
            "type": "object",
            "additionalProperties": {"type": "array", "items": {"type": "string"}},
        },
        "Tag": {
            "type": "object",
            "required": ["name"],
            "properties": {
                "name": {"type": "string"},
                "description": {"type": "string"},
                "externalDocs": {"$ref": Definitions.external_documentation},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "ExternalDocumentation": {
            "type": "object",
            "required": ["url"],
            "properties": {
                "description": {"type": "string"},
                "url": {"type": "string", "format": "uri-reference"},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "ExampleXORExamples": {
            "description": "Example and examples are mutually exclusive",
            "not": {"required": ["example", "examples"]},
        },
        "SchemaXORContent": {
            "description": "Schema and content are mutually exclusive, at least one is required",
            "not": {"required": ["schema", "content"]},
            "oneOf": [
                {"required": ["schema"]},
                {
                    "required": ["content"],
                    "description": "Some properties are not allowed if content is present",
                    "allOf": [
                        {"not": {"required": ["style"]}},
                        {"not": {"required": ["explode"]}},
                        {"not": {"required": ["allowReserved"]}},
                        {"not": {"required": ["example"]}},
                        {"not": {"required": ["examples"]}},
                    ],
                },
            ],
        },
        "Parameter": {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "in": {"type": "string"},
                "description": {"type": "string"},
                "required": {"type": "boolean", "default": False},
                "deprecated": {"type": "boolean", "default": False},
                "allowEmptyValue": {"type": "boolean", "default": False},
                "style": {"type": "string"},
                "explode": {"type": "boolean"},
                "allowReserved": {"type": "boolean", "default": False},
                "schema": {
                    "oneOf": [
                        {"$ref": Definitions.schema},
                        {"$ref": Definitions.reference},
                    ]
                },
                "content": {
                    "type": "object",
                    "additionalProperties": {"$ref": Definitions.mediatype},
                    "minProperties": 1,
                    "maxProperties": 1,
                },
                "example": {},
                "examples": {
                    "type": "object",
                    "additionalProperties": {
                        "oneOf": [
                            {"$ref": Definitions.example},
                            {"$ref": Definitions.reference},
                        ]
                    },
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
            "required": ["name", "in"],
            "allOf": [
                {"$ref": Definitions.example_xor_examples},
                {"$ref": "#/definitions/SchemaXORContent"},
                {"$ref": "#/definitions/ParameterLocation"},
            ],
        },
        "ParameterLocation": {
            "description": "Parameter location",
            "oneOf": [
                {
                    "description": "Parameter in path",
                    "required": ["required"],
                    "properties": {
                        "in": {"enum": ["path"]},
                        "style": {
                            "enum": ["matrix", "label", "simple"],
                            "default": "simple",
                        },
                        "required": {"enum": [True]},
                    },
                },
                {
                    "description": "Parameter in query",
                    "properties": {
                        "in": {"enum": ["query"]},
                        "style": {
                            "enum": [
                                "form",
                                "spaceDelimited",
                                "pipeDelimited",
                                "deepObject",
                            ],
                            "default": "form",
                        },
                    },
                },
                {
                    "description": "Parameter in header",
                    "properties": {
                        "in": {"enum": ["header"]},
                        "style": {"enum": ["simple"], "default": "simple"},
                    },
                },
                {
                    "description": "Parameter in cookie",
                    "properties": {
                        "in": {"enum": ["cookie"]},
                        "style": {"enum": ["form"], "default": "form"},
                    },
                },
            ],
        },
        "RequestBody": {
            "type": "object",
            "required": ["content"],
            "properties": {
                "description": {"type": "string"},
                "content": {
                    "type": "object",
                    "additionalProperties": {"$ref": Definitions.mediatype},
                },
                "required": {"type": "boolean", "default": False},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "SecurityScheme": {
            "oneOf": [
                {"$ref": "#/definitions/APIKeySecurityScheme"},
                {"$ref": "#/definitions/HTTPSecurityScheme"},
                {"$ref": "#/definitions/OAuth2SecurityScheme"},
                {"$ref": "#/definitions/OpenIdConnectSecurityScheme"},
            ]
        },
        "APIKeySecurityScheme": {
            "type": "object",
            "required": ["type", "name", "in"],
            "properties": {
                "type": {"type": "string", "enum": ["apiKey"]},
                "name": {"type": "string"},
                "in": {"type": "string", "enum": ["header", "query", "cookie"]},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "HTTPSecurityScheme": {
            "type": "object",
            "required": ["scheme", "type"],
            "properties": {
                "scheme": {"type": "string"},
                "bearerFormat": {"type": "string"},
                "description": {"type": "string"},
                "type": {"type": "string", "enum": ["http"]},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
            "oneOf": [
                {
                    "description": "Bearer",
                    "properties": {"scheme": {"enum": ["bearer"]}},
                },
                {
                    "description": "Non Bearer",
                    "not": {"required": ["bearerFormat"]},
                    "properties": {"scheme": {"not": {"enum": ["bearer"]}}},
                },
            ],
        },
        "OAuth2SecurityScheme": {
            "type": "object",
            "required": ["type", "flows"],
            "properties": {
                "type": {"type": "string", "enum": ["oauth2"]},
                "flows": {"$ref": "#/definitions/OAuthFlows"},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "OpenIdConnectSecurityScheme": {
            "type": "object",
            "required": ["type", "openIdConnectUrl"],
            "properties": {
                "type": {"type": "string", "enum": ["openIdConnect"]},
                "openIdConnectUrl": {"type": "string", "format": "uri-reference"},
                "description": {"type": "string"},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "OAuthFlows": {
            "type": "object",
            "properties": {
                "implicit": {"$ref": "#/definitions/ImplicitOAuthFlow"},
                "password": {"$ref": "#/definitions/PasswordOAuthFlow"},
                "clientCredentials": {"$ref": "#/definitions/ClientCredentialsFlow"},
                "authorizationCode": {
                    "$ref": "#/definitions/AuthorizationCodeOAuthFlow"
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "ImplicitOAuthFlow": {
            "type": "object",
            "required": ["authorizationUrl", "scopes"],
            "properties": {
                "authorizationUrl": {"type": "string", "format": "uri-reference"},
                "refreshUrl": {"type": "string", "format": "uri-reference"},
                "scopes": {
                    "type": "object",
                    "additionalProperties": {"type": "string"},
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "PasswordOAuthFlow": {
            "type": "object",
            "required": ["tokenUrl"],
            "properties": {
                "tokenUrl": {"type": "string", "format": "uri-reference"},
                "refreshUrl": {"type": "string", "format": "uri-reference"},
                "scopes": {
                    "type": "object",
                    "additionalProperties": {"type": "string"},
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "ClientCredentialsFlow": {
            "type": "object",
            "required": ["tokenUrl"],
            "properties": {
                "tokenUrl": {"type": "string", "format": "uri-reference"},
                "refreshUrl": {"type": "string", "format": "uri-reference"},
                "scopes": {
                    "type": "object",
                    "additionalProperties": {"type": "string"},
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "AuthorizationCodeOAuthFlow": {
            "type": "object",
            "required": ["authorizationUrl", "tokenUrl"],
            "properties": {
                "authorizationUrl": {"type": "string", "format": "uri-reference"},
                "tokenUrl": {"type": "string", "format": "uri-reference"},
                "refreshUrl": {"type": "string", "format": "uri-reference"},
                "scopes": {
                    "type": "object",
                    "additionalProperties": {"type": "string"},
                },
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
        },
        "Link": {
            "type": "object",
            "properties": {
                "operationId": {"type": "string"},
                "operationRef": {"type": "string", "format": "uri-reference"},
                "parameters": {"type": "object", "additionalProperties": {}},
                "requestBody": {},
                "description": {"type": "string"},
                "server": {"$ref": Definitions.server},
            },
            "patternProperties": {"^x-": {}},
            "additionalProperties": False,
            "not": {
                "description": "Operation Id and Operation Ref are mutually exclusive",
                "required": ["operationId", "operationRef"],
            },
        },
        "Callback": {
            "type": "object",
            "additionalProperties": {"$ref": "#/definitions/PathItem"},
            "patternProperties": {"^x-": {}},
        },
        "Encoding": {
            "type": "object",
            "properties": {
                "contentType": {"type": "string"},
                "headers": {
                    "type": "object",
                    "additionalProperties": {"$ref": Definitions.header},
                },
                "style": {
                    "type": "string",
                    "enum": ["form", "spaceDelimited", "pipeDelimited", "deepObject"],
                },
                "explode": {"type": "boolean"},
                "allowReserved": {"type": "boolean", "default": False},
            },
            "additionalProperties": False,
        },
    },
}


if __name__ == "__main__":
    import json

    print(json.dumps(SCHEMA))



import dataclasses
import typing as t

from flask.views import MethodView, View
from vbcore.http import httpcode, HttpMethod

from flaskel.ext.default import builder
from flaskel.http.exceptions import abort

UrlsType = t.Tuple[t.Union["UrlRule", str], ...]


@dataclasses.dataclass(frozen=True)
class UrlRule:
    url: str
    endpoint: t.Optional[str] = None
    options: t.Optional[dict] = None
    provide_automatic_options: t.Optional[bool] = None


class ViewSupportMixin:
    methods = [
        HttpMethod.GET,
    ]
    default_view_name: str = ""
    default_urls: UrlsType = ()

    @staticmethod
    def not_implemented() -> t.NoReturn:  # pragma: no cover
        abort(httpcode.NOT_IMPLEMENTED)

    @classmethod
    def normalize_url(cls, url: str) -> str:
        return f"/{url.lstrip('/')}"

    @classmethod
    def add_url_rule(
        cls,
        app,
        rule: UrlRule,
        view_func: t.Optional[t.Callable] = None,
    ):
        app.add_url_rule(
            cls.normalize_url(rule.url),
            view_func=view_func,
            methods=cls.methods,
            endpoint=rule.endpoint,
            provide_automatic_options=rule.provide_automatic_options,
            **(rule.options or {}),
        )

    @classmethod
    def prepare_view_func(
        cls,
        name: t.Optional[str] = None,
        view: t.Optional[t.Type["BaseView"]] = None,
        **kwargs,
    ):
        _class = view or cls
        name = name or _class.__name__
        return _class.as_view(name, **kwargs)  # type: ignore

    @classmethod
    def register(
        cls,
        app,
        name: t.Optional[str] = None,
        urls: UrlsType = (),
        view: t.Optional[t.Type["BaseView"]] = None,
        **kwargs,
    ):
        raise NotImplementedError


class BaseView(View, ViewSupportMixin):
    default_view_name: str = "index"
    default_urls: UrlsType = ("/",)

    def dispatch_request(self, *_, **__):
        """
        Must be implemented in every subclass
        """
        return self.not_implemented()  # pragma: no cover

    @classmethod
    def register(
        cls,
        app,
        name: t.Optional[str] = None,
        urls: UrlsType = (),
        view: t.Optional[t.Type["BaseView"]] = None,
        **kwargs,
    ) -> t.Callable:
        """

        :param app: Flask app or blueprint
        :param name: optional view name
        :param urls: optional urls or view name
        :param view: view class, subclass che call this must pass its reference
        :param kwargs: argument passed to cls constructor
        """
        view_func = cls.prepare_view_func(name, view, **kwargs)
        if not urls:
            cls.add_url_rule(app, UrlRule(url=name), view_func)
        for rule in urls:
            _rule = UrlRule(url=rule) if isinstance(rule, str) else rule
            cls.add_url_rule(app, _rule, view_func)
        return view_func


class Resource(MethodView, ViewSupportMixin):
    default_view_name: str = "resource"

    methods_collection = [
        HttpMethod.GET,
        HttpMethod.POST,
    ]
    methods_subresource = [
        HttpMethod.GET,
        HttpMethod.POST,
    ]
    methods_resource = [
        HttpMethod.GET,
        HttpMethod.PUT,
        HttpMethod.DELETE,
    ]

    @classmethod
    def _register_urls(
        cls, app, view_func: t.Callable, url: str, pk_type: str = "int", **kwargs
    ):
        url = cls.normalize_url(url)
        if cls.methods_collection:
            app.add_url_rule(
                url,
                view_func=view_func,
                methods=cls.methods_collection,
                **kwargs,
            )
        if cls.methods_resource:
            app.add_url_rule(
                f"{url}/<{pk_type}:res_id>",
                view_func=view_func,
                methods=cls.methods_resource,
                **kwargs,
            )
        if cls.methods_subresource:
            app.add_url_rule(
                f"{url}/<{pk_type}:res_id>/<sub_resource>",
                view_func=view_func,
                methods=cls.methods_subresource,
                **kwargs,
            )

    @classmethod
    def register(
        cls,
        app,
        name: t.Optional[str] = None,
        urls: UrlsType = (),
        view: t.Optional[t.Type[BaseView]] = None,
        **kwargs,
    ) -> t.Callable:
        view_func = cls.prepare_view_func(name, view, **kwargs)
        pk_type = kwargs.pop("pk_type", "int")  # type of res_id

        if not urls:
            cls._register_urls(app, view_func, name, pk_type)
        for rule in urls:
            if isinstance(rule, str):
                cls._register_urls(app, view_func, rule, pk_type)
            else:
                cls._register_urls(
                    app,
                    view_func,
                    rule.url,
                    pk_type,
                    endpoint=rule.endpoint,
                    provide_automatic_options=rule.provide_automatic_options,
                    **(rule.options or {}),
                )
        return view_func

    @builder.no_content
    def head(self, *args, **kwargs):
        return self.get(*args, **kwargs)

    @builder.on_accept()
    def get(self, *args, res_id=None, sub_resource=None, **kwargs):
        if res_id is None:
            return self.on_collection(*args, **kwargs)
        if sub_resource is None:
            return self.on_get(res_id, *args, **kwargs)
        if self.methods_subresource and HttpMethod.GET not in self.methods_subresource:
            abort(httpcode.METHOD_NOT_ALLOWED)

        _sub_resource = getattr(self, f"sub_{sub_resource}", None)
        if _sub_resource is None:
            abort(httpcode.NOT_FOUND)
        return _sub_resource(res_id, *args, **kwargs)

    @builder.on_accept()
    def post(self, *args, res_id=None, sub_resource=None, **kwargs):
        if res_id is None:
            return self.on_post(*args, **kwargs)
        if self.methods_subresource and HttpMethod.POST not in self.methods_subresource:
            abort(httpcode.METHOD_NOT_ALLOWED)

        _sub_resource = getattr(self, f"sub_{sub_resource}_post", None)
        if _sub_resource is None:
            abort(httpcode.NOT_FOUND)
        return _sub_resource(res_id, *args, **kwargs)

    @builder.no_content
    def delete(self, res_id, *args, **kwargs):
        return self.on_delete(res_id, *args, **kwargs)

    @builder.on_accept()
    def put(self, *args, res_id=None, **kwargs):
        return self.on_put(*args, res_id=res_id, **kwargs)

    def on_get(self, *_, **__):
        return self.not_implemented()  # pragma: no cover

    def on_post(self, *_, **__):
        return self.not_implemented()  # pragma: no cover

    def on_put(self, *_, **__):
        return self.not_implemented()  # pragma: no cover

    def on_delete(self, *_, **__):
        return self.not_implemented()  # pragma: no cover

    def on_collection(self, *_, **__):
        return self.not_implemented()  # pragma: no cover



from flask_limiter import Limiter
from vbcore.http import httpcode

import flaskel

from .default import cfremote

cap = flaskel.cap
config = flaskel.ConfigProxy("LIMITER")


def response_ok(res) -> bool:
    return httpcode.is_ok(res.status_code)


def response_ko(res) -> bool:
    return res.status_code != httpcode.TOO_MANY_REQUESTS and httpcode.is_ko(
        res.status_code
    )


limiter: Limiter = Limiter(
    key_func=cfremote.get_remote,
    default_limits=[lambda: config["FAIL"]],
    default_limits_deduct_when=response_ko,
)


def header_whitelist() -> bool:
    if not config.BYPASS_VALUE:
        return False
    return flaskel.request.headers.get(config.BYPASS_KEY) == config.BYPASS_VALUE


@limiter.request_filter
def _header_whitelist():
    return header_whitelist()


class RateLimit:
    limiter: Limiter = limiter

    @classmethod
    def slow(cls):
        return cls.limiter.limit(lambda: config.SLOW, deduct_when=response_ok)

    @classmethod
    def medium(cls):
        return cls.limiter.limit(lambda: config.MEDIUM, deduct_when=response_ok)

    @classmethod
    def fast(cls):
        return cls.limiter.limit(lambda: config.FAST, deduct_when=response_ok)

    @classmethod
    def fail(cls):
        return cls.limiter.limit(lambda: config.FAIL, deduct_when=response_ko)



import functools
import inspect
import typing as t

from vbcore.batch import BatchExecutor
from vbcore.datastruct import ObjectDict
from vbcore.http import httpcode, HttpMethod, rpc

from flaskel import abort, cap, request, Response
from flaskel.ext.default import builder

from .base import BaseView, UrlsType

# {
#     None: {
#         "action1": funct1,
#         "action2": funct2,
#         ...
#     },
#     "method1": {
#         "action1": funct1,
#         "action2": funct2,
#         ...
#     },
#     ...
# }
OperationsType = t.Dict[t.Optional[str], t.Dict[str, t.Callable]]
RPCPayloadType = t.Union[t.List[dict], dict]


class JSONRPCView(BaseView):
    version: str = "2.0"
    separator: str = "."
    operations: OperationsType = {}

    default_view_name = "jsonrpc"
    default_urls = ("/jsonrpc",)

    methods = [
        HttpMethod.POST,
    ]
    decorators = [
        builder.response("json"),
    ]

    def __init__(
        self,
        operations: OperationsType = None,
        batch_executor: t.Type[BatchExecutor] = BatchExecutor,
        **kwargs,
    ):
        self.operations = operations or {}
        self._batch_executor = batch_executor
        self._batch_args = kwargs

    def _validate_request(self, data: dict):
        if "jsonrpc" not in data or "method" not in data:
            raise rpc.RPCInvalidRequest() from None

        if data["jsonrpc"] != self.version:
            raise rpc.RPCInvalidRequest(
                f"jsonrpc version is {self.version}", req_id=data.get("id")
            ) from None

    def _validate_payload(self) -> t.Tuple[t.List[dict], bool]:
        payload: RPCPayloadType = request.get_json(True)
        if not payload:
            raise rpc.RPCParseError() from None

        if isinstance(payload, (list, tuple)):
            max_requests = cap.config.JSONRPC_BATCH_MAX_REQUEST  # type: ignore
            if max_requests and len(payload) > max_requests:
                mess = f"Operations in a single http request must be less than {max_requests}"
                abort(httpcode.REQUEST_ENTITY_TOO_LARGE, mess)
            for d in payload:
                self._validate_request(d)
        else:
            self._validate_request(payload)

        return (payload, True) if isinstance(payload, list) else ([payload], False)

    def _get_action(self, method: str) -> t.Callable:
        m = method.split(self.separator)
        if len(m) > 1:
            op, action = m[0], self.separator.join(m[1:])
        else:
            op, action = None, m[0]

        try:
            return self.operations[op][action]
        except (IndexError, TypeError, KeyError) as exc:
            cap.logger.debug(exc)
            raise rpc.RPCMethodNotFound()

    def dispatch_request(self, *_, **__):
        tasks = []
        responses = []

        try:
            payload, is_batch = self._validate_payload()
        except rpc.RPCError as ex:
            return (
                ObjectDict(
                    jsonrpc=self.version,
                    id=getattr(ex, "req_id", None),
                    error=ex.as_dict(),
                ),
                httpcode.BAD_REQUEST,
            )

        for d in payload:
            resp = ObjectDict(jsonrpc=self.version, id=None)
            try:
                params = d.get("params") or {}
                if "id" not in d:
                    tasks.append((self._get_action(d["method"]), params))
                else:
                    resp.id = d.get("id")
                    action = self._get_action(d["method"])
                    resp.result = action(**params)
            except rpc.RPCError as exc:
                cap.logger.exception(exc)
                resp.error = exc.as_dict()
            except Exception as exc:  # pylint: disable=broad-except
                cap.logger.exception(exc)
                mess = str(exc) if cap.debug is True else None
                resp.error = rpc.RPCInternalError(message=mess).as_dict()

            if "id" in d:
                responses.append(resp)

        self._batch_executor(tasks=tasks, **self._batch_args).run()
        return self.prepare_responses(responses, is_batch)

    @classmethod
    def prepare_responses(cls, responses: t.List[dict], is_batch: bool):
        if not responses:
            res = Response.no_content()
            return None, res.status_code, res.headers

        if is_batch:
            if len(responses) > 1:
                return responses, httpcode.MULTI_STATUS
            return responses

        return responses[0]

    @classmethod
    def load_from_object(cls, obj):
        for name, member in inspect.getmembers(obj, predicate=inspect.isroutine):
            if not name.startswith("_"):
                cls.method(obj.__class__.__name__, name)(member)

    @classmethod
    def method(cls, name: t.Optional[str] = None, operation: t.Optional[str] = None):
        def _method(func):
            _name = name or func.__name__

            @functools.wraps(func)
            def wrapped():
                obj = {operation: func}
                if _name not in cls.operations:
                    cls.operations[_name] = obj
                else:
                    cls.operations[_name].update(obj)

            return wrapped()

        return _method

    @classmethod
    def register(
        cls,
        app,
        name: t.Optional[str] = None,
        urls: UrlsType = (),
        view: t.Optional[t.Type[BaseView]] = None,
        **kwargs,
    ) -> t.Callable:
        kwargs.setdefault("operations", cls.operations)
        return super().register(app, name, urls, view, **kwargs)



from vbcore.http import HttpMethod

from flaskel import db_session, ExtProxy, Response
from flaskel.views import BaseView

from .repo import Payment, PaymentRepo


class PaymentView(BaseView):
    model = Payment
    repo_class = PaymentRepo
    methods = [HttpMethod.POST]
    gateway = ExtProxy("stripe")

    def __init__(self, session=None):
        self.repo = self.repo_class(session or db_session, self.model)


class PaymentIntentView(PaymentView):
    default_view_name = "payment_intent"
    default_urls = ("/stripe/payment/intent",)

    def dispatch_request(self, *_, **__):
        intent = self.gateway.create_payment_intent(self.get_amount())
        self.repo.save(intent, client_id=self.get_client_id())
        return intent

    def get_amount(self):
        """must be implemented in subclass"""
        return self.not_implemented()

    def get_client_id(self):
        """must be implemented in subclass"""
        return self.not_implemented()


class PaymentWebhookView(PaymentView):
    default_view_name = "payment_webhook"
    default_urls = ("/stripe/payment/webhook",)

    def dispatch_request(self, *_, **__):
        status, event = self.gateway.webhook_handler()
        if status is not None:
            self.repo.update(status, event)

        return Response.no_content()



import dataclasses
import typing as t

from vbcore.datastruct import ObjectDict
from vbcore.http import httpcode, HttpMethod
from vbcore.http.client import HTTPClient
from vbcore.system import CpuStat, DiskStat, MemoryStat, NetStat, SwapStat, SysStats

ConfigType = t.Optional[t.Union[t.Callable, dict]]
CheckerResponseType = t.Tuple[bool, t.Optional[t.Any]]
SuccessResponse = True, None


def health_sqlalchemy(db, app, stm: str = "SELECT 1") -> CheckerResponseType:
    try:
        with app.app_context():
            with db.engine.connect() as connection:
                connection.execute(stm)
    except Exception as exc:  # pylint: disable=broad-except
        app.logger.exception(exc)
        return False, str(exc)
    return SuccessResponse


def health_mongo(db, app, cmd: str = "ping") -> CheckerResponseType:
    try:
        with app.app_context():
            db.db.command(cmd)
    except Exception as exc:  # pylint: disable=broad-except
        app.logger.exception(exc)
        return False, str(exc)
    return SuccessResponse


def health_redis(db, app) -> CheckerResponseType:
    try:
        with app.app_context():
            db.ping()
    except Exception as exc:  # pylint: disable=broad-except
        app.logger.exception(exc)
        return False, str(exc)
    return SuccessResponse


def prepare_config(conf: ConfigType = None) -> ObjectDict:
    if not conf:
        return ObjectDict()
    return ObjectDict(**conf) if isinstance(conf, dict) else conf()


def check_mem_stat(
    threshold: float,
) -> t.Tuple[MemoryStat, t.Optional[t.Dict[str, t.Any]]]:
    stat = SysStats.memory_stats()
    if stat.percent > threshold:
        return stat, {
            "resource": "MEM",
            "usage": stat.percent,
            "threshold": threshold,
        }
    return stat, None


def check_swap_stat(
    threshold: float,
) -> t.Tuple[SwapStat, t.Optional[t.Dict[str, t.Any]]]:
    stat = SysStats.memory_stats().swap
    if stat.percent > threshold:
        return stat, {
            "resource": "SWAP",
            "usage": stat.percent,
            "threshold": threshold,
        }
    return stat, None


def check_cpu_stat(
    threshold: float,
) -> t.Tuple[CpuStat, t.Optional[t.Dict[str, t.Any]]]:
    stat = SysStats.cpu_stats()
    if stat.percent > threshold:
        return stat, {
            "resource": "CPU",
            "usage": stat.percent,
            "threshold": threshold,
        }
    return stat, None


def check_disks_stat(
    threshold: float,
    mount_points: t.Optional[t.Tuple[str, ...]] = None,
) -> t.Tuple[t.Dict[str, DiskStat], t.List[t.Dict[str, t.Any]]]:
    stats: t.Dict[str, DiskStat] = {}
    errors: t.List[t.Dict[str, t.Any]] = []

    for mount_point, stat in SysStats.disk_stats().items():
        if mount_points and mount_point not in mount_points:
            continue
        stats[mount_point] = stat
        if stat.percent > threshold:
            errors.append(
                {
                    "resource": "DISK",
                    "mount_point": mount_point,
                    "usage": stat.percent,
                    "threshold": threshold,
                }
            )
    return stats, errors


def check_nets_stat(
    threshold: float,
    nics: t.Optional[t.Tuple[str, ...]] = None,
) -> t.Tuple[t.Dict[str, NetStat], t.List[t.Dict[str, t.Any]]]:
    stats: t.Dict[str, NetStat] = {}
    errors: t.List[t.Dict[str, t.Any]] = []

    for nic, stat in SysStats.net_stats().items():
        if nics and nic not in nics:
            continue
        stats[nic] = stat
        if stat.errin > threshold or stat.errout > threshold:
            errors.append(
                {
                    "resource": "NIC",
                    "nic": nic,
                    "errors_in": stat.errin,
                    "errors_out": stat.errout,
                    "threshold": threshold,
                }
            )
    return stats, errors


def health_system(config: ConfigType = None, **__) -> CheckerResponseType:
    conf = ObjectDict(
        SYSTEM_NICS=(),
        SYSTEM_NET_ERROR_THRESHOLD=1000,
        SYSTEM_CPU_THRESHOLD=90,
        SYSTEM_MEM_THRESHOLD=90,
        SYSTEM_FS_THRESHOLD=85,
        SYSTEM_FS_MOUNT_POINTS=(),
        SYSTEM_DUMP_ALL=False,
    )
    conf.update(prepare_config(config))
    resp = ObjectDict(errors=[])

    mem_stat, mem_error = check_mem_stat(conf.SYSTEM_MEM_THRESHOLD)
    if mem_error:
        resp.errors.append(mem_error)

    swap_stat, swap_error = check_swap_stat(conf.SYSTEM_MEM_THRESHOLD)
    if swap_error:
        resp.errors.append(swap_error)

    cpu_stat, cpu_error = check_cpu_stat(conf.SYSTEM_CPU_THRESHOLD)
    if cpu_error:
        resp.errors.append(cpu_error)

    disks_stat, disks_error = check_disks_stat(
        conf.SYSTEM_FS_THRESHOLD, conf.SYSTEM_FS_MOUNT_POINTS
    )
    resp.errors.extend(disks_error or [])

    nets_stat, nets_error = check_nets_stat(
        conf.SYSTEM_NET_ERROR_THRESHOLD, conf.SYSTEM_NICS
    )
    resp.errors.extend(nets_error or [])

    resp.mem = mem_stat
    resp.swap = swap_stat
    resp.cpu = cpu_stat
    resp.disks = disks_stat
    resp.nets = nets_stat

    output = resp if conf.SYSTEM_DUMP_ALL else (resp.errors or None)
    return bool(not resp.errors), output


@dataclasses.dataclass(frozen=True)
class ServiceRequest:
    url: str
    method: str = HttpMethod.GET
    user: t.Optional[str] = None
    password: t.Optional[str] = None
    token: t.Optional[str] = None
    headers: t.Optional[dict] = None
    params: t.Optional[dict] = None
    json: t.Optional[dict] = None


def health_services(
    app, *_, conf_key: str = "HEALTH_SERVICES", **__
) -> CheckerResponseType:
    status = True
    services = app.config[conf_key]
    response: t.Dict[str, t.Dict[str, t.Any]] = {}

    for service, conf in services.items():
        request = ServiceRequest(**conf)
        client = HTTPClient(
            None,
            token=request.token,
            username=request.user,
            password=request.password,
            logger=app.logger,
        )
        res = client.request(
            uri=request.url,
            method=request.method,
            json=request.json,
            headers=request.headers,
            params=request.params,
        )

        response[service] = {"status": res.status, "message": res.body}
        if res.status != httpcode.SUCCESS:
            status = False
            response[service]["exception"] = (
                str(res.exception) if res.exception else None
            )

    return status, response



from flask import Blueprint

from flaskel.ext.default import cors, error_handler
from flaskel.views.static import SPAView

bp_api = Blueprint(
    "api",
    __name__,
    subdomain="api",
    static_folder=None,
    static_url_path=None,
    template_folder=None,
)

bp_auth = Blueprint(
    "auth",
    __name__,
    subdomain="api",
    url_prefix="/auth",
    static_folder=None,
    static_url_path=None,
    template_folder=None,
)

bp_spa = Blueprint(
    "spa",
    __name__,
    url_prefix="/",
    template_folder=SPAView.default_template_folder,
    static_folder=SPAView.default_static_folder,
    static_url_path=SPAView.default_static_url_path,
)

bp_web = Blueprint(
    "web",
    __name__,
    url_prefix="/",
    template_folder="templates",
    static_folder="static",
    static_url_path="/static/",
)


cors.init_app(bp_api)
cors.init_app(bp_auth)
error_handler.api_register(bp_api)
error_handler.api_register(bp_auth)
error_handler.web_register(bp_spa)
error_handler.web_register(bp_web)



"""
Flaskel
-------
"""
import os
import re
import sys

from pkg_resources import parse_requirements
from setuptools import find_packages as base_find_packages, setup
from setuptools.command.test import test

LICENSE = "MIT"
PLATFORMS = "any"
PYTHON_VERSION = ">3.6"
URL = "https://github.com/cs91chris/flaskel"
DESCRIPTION = "Skeleton for flask applications"

PKG_NAME = "flaskel"
PKG_TEST = "tests"
PKG_SCRIPTS = f"{PKG_NAME}.scripts"

BASE_PATH = os.path.dirname(__file__)
SKEL_DIR = os.path.join(*PKG_SCRIPTS.split("."), "skeleton")
VERSION_FILE = os.path.join(BASE_PATH, PKG_NAME, "version.py")
REQUIRES = [
    "Flask",
    "Flask-Caching",
    "Flask-CloudflareRemote",
    "Flask-Cors",
    "Flask-ErrorsHandler",
    "Flask-HTTPAuth",
    "Flask-JWT-Extended",
    "Flask-Limiter",
    "Flask-Logify",
    "Flask-ResponseBuilder",
    "Flask-SQLAlchemy",
    "Flask-TemplateSupport",
    "webargs",
    "aiohttp",
    "nest-asyncio",
    "vbcore[all]",
]
REQUIRES_EXT = REQUIRES + [
    "pyfcm",
    "stripe",
    "redis",
    "rejson",
    "hiredis",
    "Flask-Mail",
    "Flask-APScheduler",
    "flask_pymongo",
]
REQUIRES_TEST = REQUIRES_EXT + [
    "coverage",
    "pytest",
    "pytest-cov",
]

ENTRY_POINTS = dict(
    console_scripts=[
        f"{PKG_NAME}={PKG_SCRIPTS}.cli:cli",
    ]
)

try:
    # must be after setuptools
    # noinspection PyPackageRequirements
    # noinspection PyPackageRequirements,PyPep8Naming
    import Cython.Compiler.Options as cython_options
    from Cython.Build import cythonize as base_cythonize

    cython_options.docstrings = False
except ImportError:
    cython_options = None
    base_cythonize = None

if "--cythonize" not in sys.argv:
    base_cythonize = None  # pylint: disable=C0103
else:
    sys.argv.remove("--cythonize")


def ext_paths(root_dir, exclude=()):
    paths = []
    for root, _, files in os.walk(root_dir):
        for filename in files:
            file_path = os.path.join(root, filename)
            file_ext = filename.split(".")[-1]
            if file_path in exclude or file_ext not in ("py", "pyx"):
                continue

            paths.append(file_path)
    return paths


def read(file):
    with open(os.path.join(BASE_PATH, file), encoding="utf-8") as f:
        return f.read()


def grep(file, name):
    values = re.findall(rf"{name}\W*=\W*\"([^\"]+)", read(file))
    return values[0] if values else None


def readme(file):
    try:
        return read(file)
    except OSError as exc:
        print(str(exc), file=sys.stderr)
    return None


def read_requirements(filename):
    return [str(req) for req in parse_requirements(read(filename))]


def skeleton_files():
    paths = []
    for path, _, filenames in os.walk(SKEL_DIR):
        for f in filenames:
            paths.append(os.path.join(path, f))
    return paths


class PyTest(test):
    def finalize_options(self):
        test.finalize_options(self)

    def run_tests(self):
        import pytest  # pylint: disable=C0415

        sys.exit(pytest.main([PKG_TEST]))


def cythonize(paths):
    if base_cythonize is not None:
        # noinspection PyCallingNonCallable
        return base_cythonize(paths, language_level=3)
    return None


def find_packages():
    if base_cythonize is not None:
        return [PKG_SCRIPTS]
    return base_find_packages(exclude=(PKG_TEST, f"{PKG_TEST}.*"))


setup(
    name="Flaskel",
    url=URL,
    license=LICENSE,
    description=DESCRIPTION,
    platforms=PLATFORMS,
    python_requires=PYTHON_VERSION,
    long_description=readme("README.md"),
    long_description_content_type="text/markdown",
    version=grep(VERSION_FILE, "__version__"),
    author=grep(VERSION_FILE, "__author_name__"),
    author_email=grep(VERSION_FILE, "__author_email__"),
    zip_safe=False,
    include_package_data=True,
    packages=find_packages(),
    ext_modules=cythonize(ext_paths(PKG_NAME, skeleton_files())),
    test_suite=PKG_TEST,
    entry_points=ENTRY_POINTS,
    cmdclass=dict(test=PyTest),
    install_requires=REQUIRES,
    extras_require={
        "test": REQUIRES_TEST,
        "all": REQUIRES_EXT,
    },
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    ],
)



from flask_pymongo import PyMongo
from vbcore.datastruct import ObjectDict
from vbcore.date_helper import Seconds


class FlaskMongoDB(PyMongo):
    def init_app(self, app, *args, ext_name: str = "default", **kwargs):
        self.set_default_config(app, **kwargs)
        super().init_app(app, *args, **app.config.MONGO_OPTS)
        setattr(app, "extensions", getattr(app, "extensions", {}))
        app.extensions["mongo"] = ObjectDict()
        app.extensions["mongo"][ext_name] = self

    @classmethod
    def set_default_config(cls, app, **kwargs):
        app.config.setdefault("MONGO_URI", "mongodb://localhost")
        app.config.setdefault("MONGO_OPTS", {})

        timeout = app.config.MONGO_OPTS.pop("timeout", None)
        default_timeout = timeout or Seconds.millis

        app.config.MONGO_OPTS.setdefault("appname", app.config.APP_NAME)
        app.config.MONGO_OPTS.setdefault("directConnection", False)
        app.config.MONGO_OPTS.setdefault("minPoolSize", 0)
        app.config.MONGO_OPTS.setdefault("maxPoolSize", 100)
        app.config.MONGO_OPTS.setdefault("maxConnecting", 2)
        app.config.MONGO_OPTS.setdefault("retryWrites", True)
        app.config.MONGO_OPTS.setdefault("retryReads", True)
        app.config.MONGO_OPTS.setdefault("maxIdleTimeMS", None)
        app.config.MONGO_OPTS.setdefault("waitQueueTimeoutMS", None)
        app.config.MONGO_OPTS.setdefault("document_class", ObjectDict)
        app.config.MONGO_OPTS.setdefault("socketTimeoutMS", default_timeout)
        app.config.MONGO_OPTS.setdefault("connectTimeoutMS", default_timeout)
        app.config.MONGO_OPTS.setdefault("serverSelectionTimeoutMS", default_timeout)
        app.config.MONGO_OPTS.setdefault("heartbeatFrequencyMS", Seconds.millis * 10)

        app.config.MONGO_OPTS.update(**kwargs)
